from __future__ import annotations

from finesse.element import ModelElement
from PySide6 import QtCore, QtWidgets


class ModelElementRectItem(QtWidgets.QGraphicsRectItem):

    def __init__(
        self, x: float, y: float, width: float, height: float, element: ModelElement
    ):
        self.element = element
        super().__init__(x, y, width, height)
        self.setPen(QtCore.Qt.NoPen)



import sys

import finesse
from PySide6 import QtGui, QtWidgets

import virgui
from virgui.action_runner import ActionRunner
from virgui.kat_log import KatScriptLog
from virgui.katscript_viewer import KatScriptViewer
from virgui.model_layout import ModelLayout
from virgui.plotting import PlottingWidget

finesse.init_plotting()


class ZoomableGraphicsScene(QtWidgets.QGraphicsScene):

    def wheelEvent(self, event: QtWidgets.QGraphicsSceneWheelEvent) -> None:
        scale_factor = 1.15
        if event.delta() > 0:
            self.views()[0].scale(scale_factor, scale_factor)
        else:
            self.views()[0].scale(1 / scale_factor, 1 / scale_factor)
        event.accept()
        return super().wheelEvent(event)


# https://www.pythonguis.com/tutorials/pyside6-qgraphics-vector-graphics/
class Window(QtWidgets.QMainWindow):
    def __init__(self):
        super().__init__()
        self.setWindowTitle("VIRGUI")
        self.setMinimumSize(900, 900)

        self.kat_text = KatScriptViewer()
        self.kat_log = KatScriptLog()
        self.model_layout = ModelLayout(
            katscript_listener=self.kat_text.update,
            katlog_listener=self.kat_log.update,
        )
        self.plotter = PlottingWidget()
        self.action_runner = ActionRunner(plotter=self.plotter)

        self.tabs = QtWidgets.QTabWidget()
        self.setCentralWidget(self.tabs)

        # layout tab
        self.tabs.addTab(self.model_layout, "Layout")

        # calculate tab
        self.tab2 = QtWidgets.QWidget()
        self.tab2.setLayout(QtWidgets.QHBoxLayout())
        self.tabs.addTab(self.tab2, "Calculate")

        self.tab2.layout().addWidget(self.action_runner)
        self.tab2.layout().addWidget(self.plotter)

        # katscript tab
        self.tab3 = QtWidgets.QWidget()
        self.tab3.setLayout(QtWidgets.QHBoxLayout())
        self.tabs.addTab(self.tab3, "KatScript")
        self.tab3.layout().addWidget(self.kat_text)
        self.tab3.layout().addWidget(self.kat_log)


def main():
    app = QtWidgets.QApplication(sys.argv)
    app.setWindowIcon(QtGui.QIcon(str(virgui.ASSETS / "miron.png")))

    w = Window()
    w.show()

    app.exec()


if __name__ == "__main__":
    main()



import importlib.resources
from pathlib import Path

from finesse import Model

import virgui

GLOBAL_MODEL = Model()
LAYOUTS = Path(str(importlib.resources.files(virgui))) / "layouts"
ASSETS = Path(str(importlib.resources.files(virgui))) / "assets"



import pygments
from finesse.script.highlighter import KatScriptPygmentsLexer
from pygments.formatters import HtmlFormatter
from PySide6 import QtCore, QtWidgets

import virgui


class KatScriptViewer(QtWidgets.QWidget):

    def __init__(self, parent: QtWidgets.QWidget | None = None):
        super().__init__(parent=parent)
        self.setLayout(QtWidgets.QVBoxLayout())
        self.header = QtWidgets.QLabel(
            text="# Current KatScript",
            textFormat=QtCore.Qt.TextFormat.MarkdownText,
        )
        self.text_browser = QtWidgets.QTextBrowser()
        self.layout().addWidget(self.header)
        self.layout().addWidget(self.text_browser)
        self.update()

    @QtCore.Slot()
    def update(self):
        kat = virgui.GLOBAL_MODEL.unparse()
        # this line is too long to render nicely
        # maybe https://doc.qt.io/qtforpython-6/PySide6/QtWidgets/QTextEdit.html#PySide6.QtWidgets.QTextEdit.lineWrapMode
        kat = kat.replace(
            "# Items below could not be matched to original script, or were not present when the model was originally parsed.",
            "# Items below could not be matched to original script,\n# or were not present when the model was originally parsed."
            "",
        )
        html = pygments.highlight(
            kat,
            KatScriptPygmentsLexer(),
            HtmlFormatter(
                nobackground=True, noclasses=True, cssstyles="font-size: large"
            ),
        )
        self.text_browser.setHtml(html)



from finesse.parameter import Parameter
from PySide6 import QtCore, QtWidgets


class KatScriptLog(QtWidgets.QWidget):

    def __init__(self, parent: QtWidgets.QWidget | None = None):
        super().__init__(parent=parent)
        self.setLayout(QtWidgets.QVBoxLayout())
        self.header = QtWidgets.QLabel(
            text="# Change Log",
            textFormat=QtCore.Qt.TextFormat.MarkdownText,
        )
        self.log_text_edit = QtWidgets.QTextEdit()
        self.layout().addWidget(self.header)
        self.layout().addWidget(self.log_text_edit)

    @QtCore.Slot()
    def update(self, parameter: Parameter, old_val: float, new_val: float):
        self.log_text_edit.append(f"{parameter.full_name}: {old_val} -> {new_val}")



import copy
import traceback
from typing import Any

from finesse.components.general import ModelElement
from finesse.parameter import Parameter
from finesse.utilities.tables import Table
from PySide6 import QtCore
from PySide6.QtCore import Qt


# https://www.pythonguis.com/tutorials/pyside6-qtableview-modelviews-nWumpy-pandas/
class ParameterTableModel(QtCore.QAbstractTableModel):
    parameter_changed = QtCore.Signal(Parameter, float, float)

    def __init__(self, finesse_table: Table, element: ModelElement):
        super().__init__()
        self._finesse_table = finesse_table
        self._element = element

    def data(self, index, role):
        if role == Qt.DisplayRole:
            value = self._finesse_table.table[1:, :][index.row(), index.column()]
            return str(value)

    def rowCount(self, index):
        return self._finesse_table.table.shape[0] - 1

    def columnCount(self, index):
        return self._finesse_table.table.shape[1]

    def headerData(self, section, orientation, role):
        # section is the index of the column/row.
        if role == Qt.DisplayRole:
            if orientation == Qt.Horizontal:
                return str(self._finesse_table.table[0, section])

    def flags(self, index: QtCore.QModelIndex):
        # only allow the values to be edited, not the labels
        if index.column == 0:
            return Qt.ItemFlag.ItemIsSelectable | Qt.ItemFlag.ItemIsEnabled
        else:
            return (
                Qt.ItemFlag.ItemIsSelectable
                | Qt.ItemFlag.ItemIsEnabled
                | Qt.ItemFlag.ItemIsEditable
            )

    def setData(
        self,
        index: QtCore.QModelIndex | QtCore.QPersistentModelIndex,
        value: Any,
        /,
        role: int,
    ) -> bool:
        if role != Qt.ItemDataRole.EditRole:
            return super().setData(index, value, role)

        # Finesse table is in reversed order to match parameter decorator order
        p = list(reversed(self._element.parameters))[index.row()]
        # we need to create a copy
        old_val = float(copy.deepcopy(p.value))
        try:
            p.value = value
        except ValueError:
            print(f"Could not set {p} with value {value}")
            traceback.print_exc()
            return False

        self._finesse_table = self._element.parameter_table(return_str=False)
        self.dataChanged.emit(index, index, role)
        self.parameter_changed.emit(p, old_val, float(p.value))
        return True



import importlib.resources
from dataclasses import dataclass
from pathlib import Path

import finesse

# import xml.etree.ElementTree as ET
import lxml.etree

import virgui

LAYOUTS = Path(str(importlib.resources.files(virgui))) / "layouts"
PREFIX = "finesse_hitbox_"


@dataclass
class Rect:
    x: float
    y: float
    width: float
    height: float


hitbox_mapping: dict[str, Rect] = {}


def parse_layout(layout_dir: Path) -> tuple[finesse.Model, dict[str, Rect], bytes]:
    assert layout_dir.exists()
    kat_file = layout_dir / "layout.kat"
    assert kat_file.exists()
    svg_file = layout_dir / "layout.svg"
    assert svg_file.exists()

    model: finesse.Model = finesse.script.parse_file(kat_file)
    tree = lxml.etree.parse(svg_file, lxml.etree.XMLParser())
    root = tree.getroot()

    # I don't know if this viewbox thing is always added by inkscape or something more general
    svg_width = float(root.attrib["width"])
    svg_height = float(root.attrib["height"])
    viewbox = [float(val) for val in root.attrib["viewBox"].split()]
    viewbox_width = viewbox[2] - viewbox[0]
    viewbox_height = viewbox[3] - viewbox[1]
    x_scaling = svg_width / viewbox_width
    y_scaling = svg_height / viewbox_height

    # we just look for anything with the 'finesse_hitbox_<component_name> value
    # and look for x,y,width,height?
    for el in root.iter():
        hitbox = None
        for key, val in el.attrib.items():
            if val.startswith(PREFIX):
                hitbox = val
        if not hitbox:
            continue

        for attr in ("x", "y", "width", "height"):
            assert attr in el.attrib

        hitbox_mapping[hitbox.split(PREFIX)[-1]] = Rect(
            x=float(el.attrib["x"]) * x_scaling,
            y=float(el.attrib["y"]) * y_scaling,
            width=float(el.attrib["width"]) * x_scaling,
            height=float(el.attrib["height"]) * y_scaling,
        )
        parent = el.getparent()
        parent.remove(el)

    # all models have an fsig element, but should not be present in the svg
    diff = set(el for el in model.elements if el != "fsig").symmetric_difference(
        set(hitbox_mapping.keys())
    )
    if len(diff):
        raise ValueError(f"Hitbox mapping and model elements dont match: {diff}")

    # render whole svg / component layer to png? or render svg directly in QT?
    # TODO maybe check size?
    # click get coordinates and find correct element
    return model, hitbox_mapping, lxml.etree.tostring(tree)



from finesse.analysis.actions import Xaxis
from matplotlib import pyplot as plt
from PySide6 import QtCore, QtWidgets

import virgui
from virgui.plotting import PlottingWidget


class ActionRunner(QtWidgets.QWidget):

    def __init__(
        self,
        plotter: PlottingWidget,
        parent: QtWidgets.QWidget | None = None,
    ):
        super().__init__(parent=parent)
        self.plotter = plotter  # could be multiple plotters?
        self.setLayout(QtWidgets.QVBoxLayout())
        self.header = QtWidgets.QLabel(
            text="# Xaxis",
            textFormat=QtCore.Qt.TextFormat.MarkdownText,
        )
        self.header.setMaximumHeight(40)
        self.layout().addWidget(self.header)

        self.form = QtWidgets.QWidget()
        self.form_layout = QtWidgets.QFormLayout()
        self.form.setLayout(self.form_layout)
        self.layout().addWidget(self.form)

        self.parameter_dropdown = QtWidgets.QComboBox()
        self.parameter_dropdown.addItems(
            [
                p.full_name
                for p in virgui.GLOBAL_MODEL.all_parameters
                if (p.datatype is float and p.changeable_during_simulation)
            ]
        )
        self.parameter_dropdown.currentTextChanged.connect(self.on_parameter_changed)
        self.form_layout.addRow("Sweep Parameter:", self.parameter_dropdown)

        self.mode_dropdown = QtWidgets.QComboBox()
        self.mode_dropdown.addItems(["lin", "log"])
        self.form_layout.addRow("Sweep mode:", self.mode_dropdown)

        self.start_spinbox = QtWidgets.QDoubleSpinBox()
        self.start_spinbox.setRange(-1e6, 1e6)
        self.form_layout.addRow("Start:", self.start_spinbox)

        self.end_spinbox = QtWidgets.QDoubleSpinBox()
        self.end_spinbox.setRange(-1e6, 1e6)
        self.form_layout.addRow("End:", self.end_spinbox)

        self.steps_spinbox = QtWidgets.QSpinBox()
        self.steps_spinbox.setRange(0, int(1e6))
        self.form_layout.addRow("Steps:", self.steps_spinbox)

        self.run_button = QtWidgets.QPushButton()
        self.run_button.setIcon(
            self.run_button.style().standardIcon(
                QtWidgets.QStyle.StandardPixmap.SP_MediaPlay
            )
        )
        self.run_button.clicked.connect(self.run_xaxis)
        self.form_layout.addRow("Run sweep:", self.run_button)

    @QtCore.Slot()
    def run_xaxis(self):
        xaxis = Xaxis(
            self.parameter_dropdown.currentText(),
            mode=self.mode_dropdown.currentText(),
            start=self.start_spinbox.value(),
            stop=self.end_spinbox.value(),
            steps=self.steps_spinbox.value(),
        )
        sol = virgui.GLOBAL_MODEL.run(xaxis)
        sol.plot(show=False)
        self.plotter.update_figure(plt.gcf())
        print(f"ran xaxis with: {xaxis.args}")

    @QtCore.Slot()
    def on_parameter_changed(self):
        p = virgui.GLOBAL_MODEL.get(self.parameter_dropdown.currentText())
        for spinbox in (self.start_spinbox, self.end_spinbox):
            if p.units:
                spinbox.setSuffix(f" [{p.units}]")
            else:
                spinbox.setSuffix("")



from __future__ import annotations

from typing import TYPE_CHECKING, Callable

from finesse.components.general import ModelElement
from finesse.detectors import PowerDetector
from PySide6 import QtCore, QtGui, QtWidgets

from virgui.component import ModelElementRectItem
from virgui.parameter_table import ParameterTableModel
from virgui.parse_layout import parse_layout

if TYPE_CHECKING:
    pass

import virgui


class ZoomableGraphicsScene(QtWidgets.QGraphicsScene):

    def wheelEvent(self, event: QtWidgets.QGraphicsSceneWheelEvent) -> None:
        scale_factor = 1.15
        if event.delta() > 0:
            self.views()[0].scale(scale_factor, scale_factor)
        else:
            self.views()[0].scale(1 / scale_factor, 1 / scale_factor)
        event.accept()
        return super().wheelEvent(event)


class ModelLayout(QtWidgets.QWidget):

    def __init__(
        self,
        katscript_listener: Callable,
        katlog_listener: Callable,
        parent: QtWidgets.QWidget | None = None,
    ):
        super().__init__(parent=parent)
        # it would be cleaner to define these in their respective classes
        # maybe there is something like an event listener
        self.katscript_listener = katscript_listener
        self.katlog_listener = katlog_listener

        self.setLayout(QtWidgets.QHBoxLayout())
        self.scene = ZoomableGraphicsScene(
            0, 0, 600, 600, backgroundBrush=QtGui.QBrush(QtCore.Qt.white)
        )
        self.scene.selectionChanged.connect(self.on_selection)
        self.switch_layout("cavity")
        # TODO user should define detectors or the layout should?
        self.circ = virgui.GLOBAL_MODEL.add(
            PowerDetector("circ", virgui.GLOBAL_MODEL.m2.p1.o)
        )
        self.trans = virgui.GLOBAL_MODEL.add(
            PowerDetector("trans", virgui.GLOBAL_MODEL.m2.p2.o)
        )
        self.refl = virgui.GLOBAL_MODEL.add(
            PowerDetector("refl", virgui.GLOBAL_MODEL.m1.p1.o)
        )

        # info window
        self.info_vbox = QtWidgets.QVBoxLayout()
        self.table_title = QtWidgets.QLabel(
            textFormat=QtCore.Qt.TextFormat.MarkdownText,
            textInteractionFlags=QtCore.Qt.TextInteractionFlag.TextBrowserInteraction,
            openExternalLinks=True,
        )
        self.table_view = QtWidgets.QTableView()
        self.info_vbox.addWidget(self.table_title)
        self.info_vbox.addWidget(self.table_view)

        self.view = QtWidgets.QGraphicsView(self.scene)
        self.view.setRenderHint(QtGui.QPainter.Antialiasing)

        self.layout().addWidget(self.view)
        self.layout().addLayout(self.info_vbox)

    def switch_layout(self, layout_name: str):
        self.scene.clear()
        model, hitbox_mapping, svg_b_string = parse_layout(virgui.LAYOUTS / layout_name)
        virgui.GLOBAL_MODEL = model
        background_pixmap = QtGui.QPixmap()
        background_pixmap.loadFromData(svg_b_string)
        background = QtWidgets.QGraphicsPixmapItem(background_pixmap)
        self.scene.addItem(background)

        # add completely transparent rectangles as hitboxes, so users can select elements
        for comp_name, rect in hitbox_mapping.items():
            hitbox = ModelElementRectItem(
                rect.x, rect.y, rect.width, rect.height, model.get(comp_name)
            )
            self.scene.addItem(hitbox)

        for item in self.scene.items():
            if item is not background:
                item.setFlag(QtWidgets.QGraphicsItem.ItemIsSelectable)

        self.katscript_listener()
        # TODO maybe adjust the view here so everything is in focus?

    @QtCore.Slot()
    def on_selection(self):
        items = self.scene.selectedItems()
        if len(items) == 0:
            self.table_view.hide()
            self.table_title.setText("")
        elif len(items) == 1:
            item = items[0]
            assert isinstance(item, ModelElementRectItem)
            assert isinstance(item.element, ModelElement)
            el: ModelElement = item.element
            # maybe pre-create these instead of on the fly
            par_table = el.parameter_table(return_str=False)
            info_table = ParameterTableModel(par_table, el)
            info_table.dataChanged.connect(self.katscript_listener)
            info_table.parameter_changed.connect(self.katlog_listener)
            self.table_view.setModel(info_table)
            self.table_view.resizeRowsToContents()
            self.table_view.show()
            modules = el.__class__.__module__.split(".")
            doc_url = f"https://finesse.ifosim.org/docs/latest/api/{modules[1]}/{modules[2]}/{el.__class__.__module__}.{el.__class__.__name__}.html#{el.__class__.__module__}.{el.__class__.__name__}"
            self.table_title.setText(
                f"# [{el.__class__.__name__}]({doc_url}): {el.name}"
            )
        else:
            raise



from matplotlib.backends.backend_qtagg import FigureCanvas
from matplotlib.backends.backend_qtagg import NavigationToolbar2QT as NavigationToolbar
from matplotlib.figure import Figure
from PySide6 import QtWidgets


# based on https://matplotlib.org/stable/gallery/user_interfaces/embedding_in_qt_sgskip.html
class PlottingWidget(QtWidgets.QWidget):

    def __init__(self, parent: QtWidgets.QWidget | None = None):
        super().__init__(parent=parent)
        self.static_canvas = FigureCanvas(Figure())
        self.plot_toolbar = NavigationToolbar(self.static_canvas)
        self.setLayout(QtWidgets.QVBoxLayout())
        self.layout().addWidget(self.plot_toolbar)
        self.layout().addWidget(self.static_canvas)

    # this is not the correct way to do it according to the matplotlib docs
    # but finesse returns a new figure object, so I can't update the existing one
    # except by replacing it.
    def update_figure(self, fig):
        temp_static_canvas = FigureCanvas(fig)
        temp_toolbar = NavigationToolbar(temp_static_canvas)

        self.layout().replaceWidget(self.plot_toolbar, temp_toolbar)
        self.layout().replaceWidget(self.static_canvas, temp_static_canvas)

        self.static_canvas = temp_static_canvas
        self.temp_toolbar = NavigationToolbar
